@@ -5,7 +5,7 @@
  * pgpool: a language independent connection pool server for PostgreSQL
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2023	PgPool Global Development Group
+ * Copyright (c) 2003-2024	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -82,6 +82,7 @@ static bool connect_using_existing_connection(POOL_CONNECTION * frontend,
 								  POOL_CONNECTION_POOL * backend,
 								  StartupPacket *sp);
 static void check_restart_request(void);
+static void check_exit_request(void);
 static void enable_authentication_timeout(void);
 static void disable_authentication_timeout(void);
 static int	wait_for_new_connections(int *fds, SockAddr *saddr);
@@ -98,6 +99,8 @@ static bool backend_cleanup(POOL_CONNECTION * volatile *frontend, POOL_CONNECTIO
 static void child_will_go_down(int code, Datum arg);
 static int opt_sort(const void *a, const void *b);
 
+static bool unix_fds_not_isset(int* fds, int num_unix_fds, fd_set* opt);
+
 /*
  * Non 0 means SIGTERM (smart shutdown) or SIGINT (fast shutdown) has arrived
  */
@@ -298,7 +301,7 @@ do_child(int *fds)
 			pool_close(child_frontend);
 			child_frontend = NULL;
 		}
-
+		update_pooled_connection_count();
 		MemoryContextSwitchTo(TopMemoryContext);
 		FlushErrorState();
 	}
@@ -323,6 +326,7 @@ do_child(int *fds)
 		/* pgpool stop request already sent? */
 		check_stop_request();
 		check_restart_request();
+		check_exit_request();
 		accepted = 0;
 		/* Destroy session context for just in case... */
 		pool_session_context_destroy();
@@ -343,6 +347,10 @@ do_child(int *fds)
 		if (front_end_fd == RETRY)
 			continue;
 
+		set_process_status(CONNECTING);
+		/* Reset any exit if idle request even if it's pending */
+		pool_get_my_process_info()->exit_if_idle = false;
+
 		con_count = connection_count_up();
 
 		if (pool_config->reserved_connections > 0)
@@ -393,6 +401,18 @@ do_child(int *fds)
 			backend_timer_expired = 0;
 		}
 
+		/*
+		 * Check whether failover/failback is ongoing and wait for it to
+		 * finish. If it actually happened, update the private backend status
+		 * because it is possible that a backend maybe went down.
+		 */
+		if (wait_for_failover_to_finish() < 0)
+			pool_initialize_private_backend_status();
+
+		/*
+		 * Connect to backend. Also do authentication between
+		 * frontend <--> pgpool and pgpool <--> backend.
+		 */
 		backend = get_backend_connection(child_frontend);
 		if (!backend)
 		{
@@ -457,7 +477,7 @@ do_child(int *fds)
 
 		/* Mark this connection pool is not connected from frontend */
 		pool_coninfo_unset_frontend_connected(pool_get_process_context()->proc_id, pool_pool_index());
-
+		update_pooled_connection_count();
 		accepted = 0;
 		connection_count_down();
 		if (pool_config->log_disconnections)
@@ -481,6 +501,8 @@ do_child(int *fds)
 	child_exit(POOL_EXIT_NO_RESTART);
 }
 
+
+
 /* -------------------------------------------------------------------
  * private functions
  * -------------------------------------------------------------------
@@ -721,7 +743,7 @@ read_startup_packet(POOL_CONNECTION * cp)
 		case 1234:				/* cancel or SSL request */
 			/* set dummy database, user info */
 			sp->database = palloc0(1);
-			sp->user = palloc(1);
+			sp->user = palloc0(1);
 			break;
 
 		default:
@@ -825,17 +847,31 @@ connect_using_existing_connection(POOL_CONNECTION * frontend,
 			for (i = 0; i < NUM_BACKENDS; i++)
 			{
 				if (VALID_BACKEND(i))
-					if (do_command(frontend, CONNECTION(backend, i),
+				{
+					/*
+					 * We want to catch and ignore errors in do_command if a
+					 * backend is just going down right now. Otherwise
+					 * do_command raises an error and disconnects the
+					 * connection to frontend. We can safely ignore error from
+					 * "SET application_name" command if the backend goes
+					 * down.
+					 */
+					PG_TRY();
+					{
+						do_command(frontend, CONNECTION(backend, i),
 								   command_buf, MAJOR(backend),
 								   MAIN_CONNECTION(backend)->pid,
-								   MAIN_CONNECTION(backend)->key, 0) != POOL_CONTINUE)
-					{
-						ereport(ERROR,
-								(errmsg("unable to process command for backend connection"),
-								 errdetail("do_command returned DEADLOCK status")));
+								   MAIN_CONNECTION(backend)->key, 0);
 					}
+					PG_CATCH();
+					{
+						/* ignore the error message */
+						MemoryContextSwitchTo(oldContext);
+						FlushErrorState();
+					}
+					PG_END_TRY();
+				}
 			}
-
 			pool_add_param(&MAIN(backend)->params, "application_name", sp->application_name);
 			set_application_name_with_string(sp->application_name);
 		}
@@ -967,6 +1003,10 @@ found:
 	}
 }
 
+/*
+ * Copy startup packet and return it.
+ * palloc is used.
+ */
 static StartupPacket *
 StartupPacketCopy(StartupPacket *sp)
 {
@@ -996,6 +1036,10 @@ StartupPacketCopy(StartupPacket *sp)
 	return new_sp;
 }
 
+/*
+ * Create a new connection to backend.
+ * Authentication is performed if requested by backend.
+ */
 static POOL_CONNECTION_POOL * connect_backend(StartupPacket *sp, POOL_CONNECTION * frontend)
 {
 	POOL_CONNECTION_POOL *backend;
@@ -1055,6 +1099,7 @@ static POOL_CONNECTION_POOL * connect_backend(StartupPacket *sp, POOL_CONNECTION
 																ALLOCSET_DEFAULT_SIZES);
 		oldContext = MemoryContextSwitchTo(frontend_auth_cxt);
 
+		/* do authentication against backend */
 		pool_do_auth(frontend, backend);
 
 		MemoryContextSwitchTo(oldContext);
@@ -1220,6 +1265,9 @@ static RETSIGTYPE authentication_timeout(int sig)
 	child_exit(POOL_EXIT_AND_RESTART);
 }
 
+/*
+ * Enable authtentication timeout.
+ */
 static void
 enable_authentication_timeout(void)
 {
@@ -1229,6 +1277,9 @@ enable_authentication_timeout(void)
 	alarm_enabled = true;
 }
 
+/*
+ * Disable authtentication timeout.
+ */
 static void
 disable_authentication_timeout(void)
 {
@@ -1239,7 +1290,9 @@ disable_authentication_timeout(void)
 	}
 }
 
-
+/*
+ * Send parameter status message to frontend.
+ */
 static void
 send_params(POOL_CONNECTION * frontend, POOL_CONNECTION_POOL * backend)
 {
@@ -1305,6 +1358,10 @@ child_will_go_down(int code, Datum arg)
 	if (pool_connection_pool)
 		close_all_backend_connections();
 }
+
+/*
+ * Exit child process. Can be called from signal handlers.
+ */
 void
 child_exit(int code)
 {
@@ -1433,6 +1490,23 @@ pool_initialize_private_backend_status(void)
 	my_main_node_id = REAL_MAIN_NODE_ID;
 }
 
+static void
+check_exit_request(void)
+{
+	/*
+	 * Check if exit request is set because of spare children management.
+	 */
+	if (pool_get_my_process_info()->exit_if_idle)
+	{
+		ereport(LOG,
+				(errmsg("Exit flag set"),
+				 errdetail("Exiting myself")));
+
+		pool_get_my_process_info()->exit_if_idle = 0;
+		child_exit(POOL_EXIT_NO_RESTART);
+	}
+}
+
 static void
 check_restart_request(void)
 {
@@ -1446,7 +1520,7 @@ check_restart_request(void)
 				(errmsg("failover or failback event detected"),
 				 errdetail("restarting myself")));
 
-		pool_get_my_process_info()->need_to_restart = 0;
+		pool_get_my_process_info()->need_to_restart = false;
 		child_exit(POOL_EXIT_AND_RESTART);
 	}
 }
@@ -1457,7 +1531,6 @@ check_restart_request(void)
  * to connect, on successful connection returns the socket descriptor
  * and returns -1 if timeout has occurred
  */
-
 static int
 wait_for_new_connections(int *fds, SockAddr *saddr)
 {
@@ -1636,6 +1709,11 @@ retry_accept:
 		pause();
 	}
 
+	/* wait for failover/failback to finish */
+	if (wait_for_failover_to_finish() < 0)
+		/* failover/failback occurred. Update private backend status */
+		pool_initialize_private_backend_status();
+
 	afd = accept(fd, (struct sockaddr *) &saddr->addr, &saddr->salen);
 
 	save_errno = errno;
@@ -1667,7 +1745,7 @@ retry_accept:
 	 * Set no delay if AF_INET socket. Not sure if this is really necessary
 	 * but PostgreSQL does this.
 	 */
-	if (!FD_ISSET(fds[0], &rmask))	/* fds[0] is UNIX domain socket */
+	if (unix_fds_not_isset(fds, pool_config->num_unix_socket_directories, &rmask))
 	{
 		on = 1;
 		if (setsockopt(afd, IPPROTO_TCP, TCP_NODELAY,
@@ -1698,6 +1776,20 @@ retry_accept:
 	return afd;
 }
 
+static bool
+unix_fds_not_isset(int* fds, int num_unix_fds, fd_set* opt)
+{
+	int		i;
+	for (i = 0; i < num_unix_fds; i++)
+	{
+		if (!FD_ISSET(fds[i], opt))
+			continue;
+
+		return false;
+	}
+	return true;
+}
+
 static void
 check_config_reload(void)
 {
@@ -1869,6 +1961,10 @@ get_connection(int front_end_fd, SockAddr *saddr)
 	return cp;
 }
 
+/*
+ * Connect to backend. Also do authentication between client <--> pgpool and
+ * pgpool <--> backend.
+ */
 static POOL_CONNECTION_POOL *
 get_backend_connection(POOL_CONNECTION * frontend)
 {
@@ -1930,6 +2026,9 @@ retry_startup:
 										ALLOCSET_DEFAULT_SIZES);
 		MemoryContext oldContext = MemoryContextSwitchTo(frontend_auth_cxt);
 
+		/*
+		 * Do frontend <-> pgpool authentication based on pool_hba.conf.
+		 */
 		ClientAuthentication(frontend);
 
 		MemoryContextSwitchTo(oldContext);
@@ -2006,7 +2105,10 @@ retry_startup:
 
 	if (backend == NULL)
 	{
-		/* create a new connection to backend */
+		/*
+		 * Create a new connection to backend.
+		 * Authentication is performed if requested by backend.
+		 */
 		backend = connect_backend(sp, frontend);
 	}
 	else
