@@ -309,7 +309,8 @@ select_load_balancing_node(void)
 				r;
 	int			i;
 	int			index_db = -1,
-				index_app = -1;
+				index_app = -1,
+				index_user = -1;
 	POOL_SESSION_CONTEXT *ses = pool_get_session_context(false);
 	int			tmp;
 	int			no_load_balance_node_id = -2;
@@ -329,6 +330,33 @@ select_load_balancing_node(void)
 	r = (((double) random()) / RAND_MAX);
 #endif
 
+	/*
+	 * Check user_redirect_preference_list
+	 */
+	if (SL_MODE && pool_config->redirect_usernames)
+	{
+		char	   *user = MAIN_CONNECTION(ses->backend)->sp->user;
+
+		/*
+		 * Check to see if the user matches any of
+		 * user_redirect_preference_list
+		 */
+		index_user = regex_array_match(pool_config->redirect_usernames, user);
+		if (index_user >= 0)
+		{
+			/* Matches */
+			ereport(DEBUG1,
+					(errmsg("selecting load balance node user matched"),
+					 errdetail("username: %s index is %d dbnode is %s weight is %f", user, index_user,
+							   pool_config->user_redirect_tokens->token[index_user].right_token,
+							   pool_config->user_redirect_tokens->token[index_user].weight_token)));
+
+			tmp = choose_db_node_id(pool_config->user_redirect_tokens->token[index_user].right_token);
+			if (tmp == -1 || (tmp >= 0 && VALID_BACKEND_RAW(tmp)))
+				suggested_node_id = tmp;
+		}
+	}
+
 	/*
 	 * Check database_redirect_preference_list
 	 */
@@ -343,6 +371,13 @@ select_load_balancing_node(void)
 		index_db = regex_array_match(pool_config->redirect_dbnames, database);
 		if (index_db >= 0)
 		{
+			/*
+			 * if the database name matches any of
+			 * database_redirect_preference_list,
+			 * user_redirect_preference_list will be ignored.
+			 */
+			index_user = -1;
+
 			/* Matches */
 			ereport(DEBUG1,
 					(errmsg("selecting load balance node db matched"),
@@ -382,11 +417,12 @@ select_load_balancing_node(void)
 				 * app_name_redirect_preference_list,
 				 * database_redirect_preference_list will be ignored.
 				 */
+				index_user = -1;
 				index_db = -1;
 
 				/* Matches */
 				ereport(DEBUG1,
-						(errmsg("selecting load balance node db matched"),
+						(errmsg("selecting load balance node app name matched"),
 						 errdetail("app_name: %s index is %d dbnode is %s weight is %f", app_name, index_app,
 								   pool_config->app_name_redirect_tokens->token[index_app].right_token,
 								   pool_config->app_name_redirect_tokens->token[index_app].weight_token)));
@@ -405,11 +441,9 @@ select_load_balancing_node(void)
 		 * and prefer_lower_delay_standby are true, we choose the least delayed
 		 * node if suggested_node is standby and delayed over delay_threshold.
 		 */
-		if (STREAM &&
-			pool_config->delay_threshold &&
-			pool_config->prefer_lower_delay_standby &&
-			(suggested_node_id != PRIMARY_NODE_ID) &&
-			(BACKEND_INFO(suggested_node_id).standby_delay > pool_config->delay_threshold))
+		if (STREAM && pool_config->prefer_lower_delay_standby &&
+			suggested_node_id != PRIMARY_NODE_ID &&
+			check_replication_delay(suggested_node_id) < 0)
 		{
 			ereport(DEBUG1,
 				(errmsg("selecting load balance node"),
@@ -419,7 +453,10 @@ select_load_balancing_node(void)
 			 * The new load balancing node is seleted from the
 			 * nodes which have the lowest delay.
 			 */
-			lowest_delay = pool_config->delay_threshold;
+			if (pool_config->delay_threshold_by_time > 0)
+				lowest_delay = pool_config->delay_threshold_by_time * 1000;	/* convert from milli seconds to micro seconds */
+			else
+				lowest_delay = pool_config->delay_threshold;
 
 			/* Initialize */
 			total_weight = 0.0;
@@ -489,7 +526,8 @@ select_load_balancing_node(void)
 		 * choose load balance node from other nodes.
 		 */
 		if ((index_db >= 0 && r <= pool_config->db_redirect_tokens->token[index_db].weight_token) ||
-			(index_app >= 0 && r <= pool_config->app_name_redirect_tokens->token[index_app].weight_token))
+			(index_app >= 0 && r <= pool_config->app_name_redirect_tokens->token[index_app].weight_token) ||
+			(index_user >= 0 && r <= pool_config->user_redirect_tokens->token[index_user].weight_token))
 		{
 			ereport(DEBUG1,
 					(errmsg("selecting load balance node"),
@@ -505,7 +543,8 @@ select_load_balancing_node(void)
 	{
 		/* If the weight is less than random rate then send to primary. */
 		if ((index_db >= 0 && r > pool_config->db_redirect_tokens->token[index_db].weight_token) ||
-			(index_app >= 0 && r > pool_config->app_name_redirect_tokens->token[index_app].weight_token))
+			(index_app >= 0 && r > pool_config->app_name_redirect_tokens->token[index_app].weight_token) ||
+			(index_user >= 0 && r > pool_config->user_redirect_tokens->token[index_user].weight_token))
 		{
 			ereport(DEBUG1,
 					(errmsg("selecting load balance node"),
@@ -561,16 +600,17 @@ select_load_balancing_node(void)
 	 * prefer_lower_delay_standby is true, we elect the most lower delayed
 	 * node if suggested_node is standby and delayed over delay_threshold.
 	 */
-	if (STREAM &&
-		pool_config->delay_threshold &&
-		pool_config->prefer_lower_delay_standby &&
-		(BACKEND_INFO(selected_slot).standby_delay > pool_config->delay_threshold))
+	if (STREAM && pool_config->prefer_lower_delay_standby &&
+		check_replication_delay(selected_slot) < 0)
 	{
 		ereport(DEBUG1,
 				(errmsg("selecting load balance node"),
 				 errdetail("backend id %d is streaming delayed over delay_threshold", selected_slot)));
 
-		lowest_delay = pool_config->delay_threshold;
+		if (pool_config->delay_threshold_by_time > 0)
+			lowest_delay = pool_config->delay_threshold_by_time * 1000;
+		else
+			lowest_delay = pool_config->delay_threshold;
 		total_weight = 0.0;
 		for (i = 0; i < NUM_BACKENDS; i++)
 		{
@@ -1053,3 +1093,40 @@ si_commit_done(void)
 		session->si_state = SI_NO_SNAPSHOT;
 	}
 }
+
+/*
+ * Check replication delay and returns the status.
+ * Return values:
+ * 0: no delay or not in streaming repplication mode or
+ * delay_threshold(_by_time) is set to 0
+ * -1: delay exceeds delay_threshold_by_time
+ * -2: delay exceeds delay_threshold
+ */
+int	check_replication_delay(int node_id)
+{
+	BackendInfo *bkinfo;
+
+	if (!STREAM)
+		return 0;
+
+	bkinfo = pool_get_node_info(node_id);
+
+	/*
+	 * Check delay_threshold_by_time.  bkinfo->standby_delay is in
+	 * microseconds while delay_threshold_by_time is in milliseconds. We need
+	 * to multiply delay_threshold_by_time by 1000 to normalize.
+	 */
+	if (pool_config->delay_threshold_by_time > 0 &&
+		bkinfo->standby_delay > pool_config->delay_threshold_by_time*1000)
+		return -1;
+
+	/*
+	 * Check delay_threshold.
+	 */
+	if (pool_config->delay_threshold > 0 &&
+		bkinfo->standby_delay > pool_config->delay_threshold)
+		return -2;
+
+	return 0;
+}
+
