@@ -16,7 +16,7 @@ function wait_for_pgpool_startup {
 }
 
 #-------------------------------------------
-# wait for primary/master failover done
+# wait for primary/main failover done
 #-------------------------------------------
 function wait_for_failover_done {
 	timeout=20
@@ -25,7 +25,7 @@ function wait_for_failover_done {
 	do
 		$PGBIN/psql -p $PGPOOL_PORT -c "show pool_nodes" test >/dev/null 2>&1
 		if [ $? = 0 ];then
-		    $PGBIN/psql -p $PGPOOL_PORT -c "show pool_nodes" test |egrep -i "primary|master">/dev/null 2>&1
+		    $PGBIN/psql -p $PGPOOL_PORT -c "show pool_nodes" test |egrep -i "primary|main">/dev/null 2>&1
 		    if [ $? = 0 ];then
 			break;
 		    fi
@@ -35,3 +35,19 @@ function wait_for_failover_done {
 		sleep 1
 	done
 }
+
+#-------------------------------------------
+# clean remaining processes and sockets
+#-------------------------------------------
+function clean_all {
+	pgrep pgpool | xargs kill -9 > /dev/null 2>&1
+	pgrep postgres | xargs kill -9 > /dev/null 2>&1
+	rm -f $PGSOCKET_DIR/.s.PGSQL.*
+}
+
+#-------------------------------------------
+# check segmentation fault
+#-------------------------------------------
+function check_segfault {
+	egrep -i "segmentation fault|segfault" log/pgpool.log >/dev/null 2>&1
+}
